/**
 *
 * Model can be added manually using the Model interface or from SVG using the `parseSVG(svg, { scale, zIndexFixed })` method.
 *
 * `scale` is a value used to scale the model. Each object geometry point is multiply by this value. (default to 1)
 * `zIndexFixed` is used to reduce the zIndex precision. To much precision reduce the viewer2D performance. (default to 3, meanings 3 decimals precision)
 *
 * Here is an example of compatible SVG:
 *
 * ```html
 * <?xml version="1.0" encoding="UTF-8"?>
 * <svg viewBox="-10, -10, 10, 10" xmlns="http://www.w3.org/2000/svg">
 *   <g id="1" area="1" elevation-top="1" ifc-type="type">
 *     <path d="m-10 0 10 0" />
 *   </g>
 * </svg>
 * ```
 *
 * `id` => object.uuid
 * `elevation-top` => object.zIndex
 * `ifc-type` => object.type
 * `area` => object.area
 *
 * Path commands supported are: m, M, h, H, v, V, l, L and Z.
 */

/**
 * @param { SVGElement } svg
 * @param { { scale: number, zIndexFixed: number } } param1
 *
 * @returns { E2D.Model }
 */
function parseSVG(svg, { scale = 1, zIndexFixed = 3 } = {}) {
  try {
    const model = {
      objects: Array.from(svg.firstChild.getElementsByTagName("g")).map(g => {
        const object = {
          uuid: g.getAttribute("id"),
          area: g.getAttribute("area"),
          zIndex: Number(g.getAttribute("elevation-top")).toFixed(zIndexFixed),
          type: g.getAttribute("ifc-type"),
        };

        const doorData = g.getAttribute("ifc-door-data");
        if (doorData) {
          object.userData = {
            doorData,
          };
        }

        const lines = parseSvgPath(g.firstElementChild.getAttribute("d"));

        object.geometry = lines.map(line =>
          line.points?.map((_, i, arr) => (arr[i] *= scale))
        );

        return object;
      }),
    };

    return model;
  } catch (err) {
    throw new Error(
      `Impossible to process svg. The format is not compatible with this loader. Error : ${err}`
    );
  }
}

function parseSvgPath(svgPath) {
  const commands = parsePath(svgPath);
  const lines = [];

  fromSvgCommandsToLines(commands, lines);

  return lines.map(line => ({ points: line }));
}

function fromSvgCommandsToLines(commands, lines) {
  let x = 0;
  let y = 0;

  let currentLine = [];

  for (let i = 0; i < commands.length; i++) {
    const command = commands[i];

    switch (command.code) {
      case "m": {
        if (currentLine.length) {
          lines.push(currentLine);
          currentLine = [];
        }

        x += command.end.x;
        y += command.end.y;
        currentLine.push(x, y);
        break;
      }
      case "M": {
        if (currentLine.length) {
          lines.push(currentLine);
          currentLine = [];
        }
        x = command.end.x;
        y = command.end.y;
        currentLine.push(x, y);
        break;
      }
      case "H": {
        const lastYIndex = currentLine.length - 1;
        x = command.value;
        const lastY = currentLine[lastYIndex];
        currentLine.push(x, lastY);
        break;
      }
      case "h": {
        const lastYIndex = currentLine.length - 1;
        x += command.value;
        const lastY = currentLine[lastYIndex];
        currentLine.push(x, lastY);
        break;
      }
      case "V": {
        const lastXIndex = currentLine.length - 2;
        y = command.value;
        const lastX = currentLine[lastXIndex];
        currentLine.push(lastX, y);
        break;
      }
      case "v": {
        const lastXIndex = currentLine.length - 2;
        y += command.value;
        const lastX = currentLine[lastXIndex];
        currentLine.push(lastX, y);
        break;
      }
      case "L": {
        x = command.end.x;
        y = command.end.y;

        currentLine.push(x, y);
        break;
      }
      case "l": {
        x += command.end.x;
        y += command.end.y;

        currentLine.push(x, y);
        break;
      }
      case "Z": {
        lines.push(currentLine);
        currentLine = [];
        break;
      }
      default: {
        throw new Error(
          "VIEWER 2D - Draw command not supported: ",
          command.code,
          command
        );
      }
    }
  }

  if (currentLine.length) {
    lines.push(currentLine);
  }
}

/*!
 * d-path-parser - v1.0.0
 * by Massimo Artizzu (MaxArt2501)
 *
 * https://github.com/MaxArt2501/d-path-parser
 *
 * Licensed under the MIT License
 * See LICENSE for details
 */
function parsePath(d) {
  var re = {
    command: /\s*([achlmqstvz])/gi,
    number: /\s*([+-]?\d*\.?\d+(?:e[+-]?\d+)?)/gi,
    comma: /\s*(?:(,)|\s)/g,
    flag: /\s*([01])/g,
  };
  var matchers = {
    number: function (must) {
      return +get("number", must);
    },
    "coordinate pair": function (must) {
      var x = get("number", must);
      if (x === null && !must) return null;
      get("comma");
      var y = get("number", true);
      return { x: +x, y: +y };
    },
    "arc definition": function (must) {
      var radii = matchers["coordinate pair"](must);
      if (!radii && !must) return null;
      get("comma");
      var rotation = +get("number", true);
      get("comma", true);
      var large = !!+get("flag", true);
      get("comma");
      var clockwise = !!+get("flag", true);
      get("comma");
      var end = matchers["coordinate pair"](true);
      return {
        radii: radii,
        rotation: rotation,
        large: large,
        clockwise: clockwise,
        end: end,
      };
    },
  };
  var index = 0;
  var commands = [];

  while (index < d.length) {
    var cmd = get("command");
    var upcmd = cmd.toUpperCase();
    var relative = cmd !== upcmd;
    var sequence;
    switch (upcmd) {
      case "M":
        sequence = getSequence("coordinate pair").map(function (coords, i) {
          if (i === 1) cmd = relative ? "l" : "L";
          return makeCommand({ end: coords });
        });
        break;
      case "L":
      case "T":
        sequence = getSequence("coordinate pair").map(function (coords) {
          return makeCommand({ end: coords });
        });
        break;
      case "C":
        sequence = getSequence("coordinate pair");
        if (sequence.length % 3)
          throw Error("Expected coordinate pair triplet at position " + index);

        sequence = sequence.reduce(function (seq, coords, i) {
          var rest = i % 3;
          if (!rest) {
            seq.push(makeCommand({ cp1: coords }));
          } else {
            var last = seq[seq.length - 1];
            last[rest === 1 ? "cp2" : "end"] = coords;
          }
          return seq;
        }, []);

        break;
      case "Q":
      case "S":
        sequence = getSequence("coordinate pair");
        if (sequence.length & 1)
          throw Error("Expected coordinate pair couple at position " + index);

        sequence = sequence.reduce(function (seq, coords, i) {
          var odd = i & 1;
          if (!odd) {
            seq.push(makeCommand({ cp: coords }));
          } else {
            var last = seq[seq.length - 1];
            last.end = coords;
          }
          return seq;
        }, []);

        break;
      case "H":
      case "V":
        sequence = getSequence("number").map(function (value) {
          return makeCommand({ value: value });
        });
        break;
      case "A":
        sequence = getSequence("arc definition").map(makeCommand);
        break;
      case "Z":
        sequence = [{ code: "Z" }];
        break;
    }
    commands.push.apply(commands, sequence);
  }

  return commands;

  function makeCommand(obj) {
    obj.code = cmd;
    obj.relative = relative;

    return obj;
  }
  function get(what, must) {
    re[what].lastIndex = index;
    var res = re[what].exec(d);
    if (!res || res.index !== index) {
      if (!must) return null;
      throw Error("Expected " + what + " at position " + index);
    }

    index = re[what].lastIndex;

    return res[1];
  }
  function getSequence(what) {
    var sequence = [];
    var matched;
    var must = true;
    while ((matched = matchers[what](must))) {
      sequence.push(matched);
      must = !!get("comma");
    }

    return sequence;
  }
}

export default parseSVG;
